/*
 Alarm System v3.0 Door
 Version: Test
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 1/04/2017
 Last Modified: 1/06/2017
 
 */

#define DELAY_STD 0
#define DELAY_WDT 1

#define PACKET_LENGTH 0

#define FUNCTION_RESET 0x00
#define FUNCTION_SLEEP 0x01
#define FUNCTION_IDLE 0x02
#define FUNCTION_COMMAND 0x03

#define COMMAND_READ 0x02
#define COMMAND_WRITE 0x12
#define COMMAND_HMAC 0x11
#define COMMAND_NONCE 0x16
#define COMMAND_RANDOM 0x1B
#define COMMAND_LOCK 0x17

#define ZONE_ENCODING_CONFIG 0
#define ZONE_ENCODING_OTP 1
#define ZONE_ENCODING_DATA 2

#define ZONE_LOCK_CONFIG 0
#define ZONE_LOCK_DATA_OTP 1

#define NONCE_PASS_THROUGH 0x03

#define READ_4_BYTES 0x00
#define READ_32_BYTES 0x80

#define LOCK_NO_CRC_CHECK 0x80

#define SLOT_0 0x00

// Calculate the CRC
void sha204c_calculate_crc(uint8_t length, uint8_t *data, uint8_t *crc) {
	uint8_t counter;
	uint16_t crc_register = 0;
	uint16_t polynom = 0x8005;
	uint8_t shift_register;
	uint8_t data_bit, crc_bit;

	for (counter = 0; counter < length; counter++) {
		for (shift_register = 0x01; shift_register > 0x00; shift_register <<= 1) {
			data_bit = (data[counter] & shift_register) ? 1 : 0;
			crc_bit = crc_register >> 15;
			crc_register <<= 1;
			if (data_bit != crc_bit) {
				crc_register ^= polynom;
			}
		}
	}
	crc[0] = (uint8_t) (crc_register & 0x00FF);
	crc[1] = (uint8_t) (crc_register >> 8);
}

// Check the CRC
uint8_t sha204c_check_crc(uint8_t *data) {
	uint8_t crc[2];

	uint8_t count = data[PACKET_LENGTH];
	sha204c_calculate_crc(count - 2, data, crc);

	if (data[(count-2)] == crc[0] && (data[(count-1)] == crc[1])) {
		return true;
	}
	return false;
}

// ATSHA204A sleep (loses TempKey data)
void atsha204a_sleep(void) {
  soft_i2c_master_start(ATSHA204A_ADDR | I2C_WRITE);
  soft_i2c_master_write(FUNCTION_SLEEP); 
  soft_i2c_master_stop();
}

// ATSHA204A idle (retains TempKey data)
void atsha204a_idle(void) {
  soft_i2c_master_start(ATSHA204A_ADDR | I2C_WRITE);
  soft_i2c_master_write(FUNCTION_IDLE); 
  soft_i2c_master_stop();
}

// Read data in the ATSHA204A buffer
uint8_t atsha204a_read_buffer(void) {
	if (soft_i2c_master_start(ATSHA204A_ADDR | I2C_READ)) {
		uint8_t x = 0;
		
		// Length
		dataBuffer[x] = soft_i2c_master_read(0);
		x++;
		
		// Read until packet length minus 2 (for CRC)
		while (x < (dataBuffer[PACKET_LENGTH] - 2)) {
			dataBuffer[x] = soft_i2c_master_read(0);
			x++;
		}
		
		// CRC
		dataBuffer[x] = soft_i2c_master_read(0);
		x++;
		dataBuffer[x] = soft_i2c_master_read(1);
		x++;
		
		soft_i2c_master_stop();
		
		// Check CRC matches
		if (sha204c_check_crc(dataBuffer) == true) {
			return true;
		}
	}
	
	return false;
}

// Wake up ATSHA204A
void atsha204a_wakeup(void) {
  DDRA |= (1<<TWI_SDA_PIN); // Pull-up taken low
  _delay_ms(1);
  DDRA &= ~(1<<TWI_SDA_PIN); // Pull-up left high
  _delay_ms(3);

  // Read wake packet
  atsha204a_read_buffer();
}

// Request random number
void atsha204a_command_random_number(uint8_t* outputBuffer, uint8_t delayType) {
	atsha204a_wakeup();
	
	uint8_t dataSend[6];
	if (soft_i2c_master_start(ATSHA204A_ADDR | I2C_WRITE)) {
		dataSend[0] = FUNCTION_COMMAND; // command
		dataSend[1] = 7; // count (included in count)
		dataSend[2] = COMMAND_RANDOM;
		dataSend[3] = 0;
		dataSend[4] = 0;
		dataSend[5] = 0;
		
		sha204c_calculate_crc(5, &dataSend[1], crc); // crc starts at count
		
		// Send
		soft_i2c_master_start(ATSHA204A_ADDR | I2C_WRITE);
		for (uint8_t x = 0; x < 6; x++) {
			soft_i2c_master_write(dataSend[x]);    
		}
		soft_i2c_master_write(crc[0]); 
		soft_i2c_master_write(crc[1]);
		soft_i2c_master_stop();
		
		if (delayType == DELAY_WDT) {
			watchdog_sleep(T64MS);
		}
		else {
			_delay_ms(50);
		}
		
		atsha204a_read_buffer();
		atsha204a_sleep();
		
		// Copy from the buffer to the output buffer specified
		for (uint8_t x = 0; x < 32; x++) {
			outputBuffer[x] = dataBuffer[x+1]; // Skip the count (first byte [0])
		}
	} 
}

// Load the random number into TempKey
void atsha204a_command_nonce(uint8_t* inputBuffer, uint8_t delayType) {
	atsha204a_wakeup();
	
	uint8_t dataSend[40];
	if (soft_i2c_master_start(ATSHA204A_ADDR | I2C_WRITE)) {
		dataSend[0] = FUNCTION_COMMAND; // command
		dataSend[1] = 7 + 32; // count (included in count)
		dataSend[2] = COMMAND_NONCE;
		dataSend[3] = NONCE_PASS_THROUGH;
		dataSend[4] = 0;
		dataSend[5] = 0;
		
		// Copy random number in
		for (uint8_t x = 0; x < 32; x++) {
			dataSend[x+6] = inputBuffer[x];
		}
		
		sha204c_calculate_crc(5 + 32, &dataSend[1], crc); // crc starts at count
		
		// Send
		soft_i2c_master_start(ATSHA204A_ADDR | I2C_WRITE);
		for (uint8_t x = 0; x < 6 + 32; x++) {
			soft_i2c_master_write(dataSend[x]);    
		}
		soft_i2c_master_write(crc[0]); 
		soft_i2c_master_write(crc[1]);
		soft_i2c_master_stop();
		
		if (delayType == DELAY_WDT) {
			watchdog_sleep(T64MS);
		}
		else {
			_delay_ms(60);
		}
		
		atsha204a_read_buffer();
		atsha204a_idle();
	}
}

// Generate HMAC from the nounce
void atsha204a_command_hmac(uint8_t delayType) {
	atsha204a_wakeup();
	
	uint8_t dataSend[6];
	if (soft_i2c_master_start(ATSHA204A_ADDR | I2C_WRITE)) {
		dataSend[0] = FUNCTION_COMMAND; // command
		dataSend[1] = 7; // count (included in count)
		dataSend[2] = COMMAND_HMAC;
		dataSend[3] = 0x04; //mode, TempKey.SourceFlag, 1 = Input seed only
		dataSend[4] = SLOT_0;
		dataSend[5] = 0;
		
		sha204c_calculate_crc(5, &dataSend[1], crc); // crc starts at count
		
		// Send
		soft_i2c_master_start(ATSHA204A_ADDR | I2C_WRITE);
		for (uint8_t x = 0; x < 6; x++) {
			soft_i2c_master_write(dataSend[x]);    
		}
		soft_i2c_master_write(crc[0]); 
		soft_i2c_master_write(crc[1]);
		soft_i2c_master_stop();
		
		if (delayType == DELAY_WDT) {
			watchdog_sleep(T64MS);
			watchdog_sleep(T16MS);
		}
		else {
			_delay_ms(70);
		}
		
		atsha204a_read_buffer();
		atsha204a_sleep();
	}
}
