/*
 Alarm System v3.0 PIR
 Version: Test
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 1/04/2017
 Last Modified: 26/06/2017
 
 */

// Check battery voltage
void check_battery_voltage(void) {
	ADCSRA |= (1<<ADEN); // Turn on ADC
	batteryVoltage = adc_read(batteryAdcPin) + adcCalibration;
	ADCSRA &= ~(1<<ADEN); // Turn off ADC
	
	// If battery voltage below 3.0V then turn off the sensor
	if (batteryVoltage <= lowVoltageLifeBat) {
		PCMSK1 &= ~(1<<reed_INT); // Remove interrup
		systemState = SYSTEM_DISABLED;
		sensorTriggered = false;
		sensorDisabled = true; // Disable sensor permanently
	}
}

// Clear data in
void clear_data_in(void) {
	for (uint8_t x = 0; x < 40; x++) {
		dataIn[x] = 0;
	}
}

// Wait for a valid packet
uint8_t wait_for_valid_packet(uint8_t waitTime) {
	RH_RF22_resetFifos();
	RH_RF22_clearInterrupts();
	RH_RF22_setModeRx();
	PCMSK0 |= (1<<nIRQ_INT); // Enable interrupt for nIRQ pin
	
	// Wait for a packet that matches our address, we will wake up from the nIRQ pin or the watchdog timer
	if (waitTime == WAIT_32MS) {
		watchdog_sleep(T32MS);
	}
	else if (waitTime == WAIT_64MS) {
		watchdog_sleep(T64MS);
	}
	PCMSK0 &= ~(1<<nIRQ_INT); // Disable interrupt
	RH_RF22_setModeStandby(); // Set to standby as quick as possible
	
	// Check if any packets have been received
	uint8_t readReg = spi_read_reg(RH_RF22_REG_03_INTERRUPT_STATUS1);
	if (!(readReg & ipkvalid)) { // No packet received to our address
		
		return false;
	}
	
	clear_data_in(); // Clear existing data
	
	// Read from address and type
	RH_RF22_fromAddress = spi_read_reg(RH_RF22_REG_48_RECEIVED_HEADER2);
	RH_RF22_requestType = spi_read_reg(RH_RF22_REG_49_RECEIVED_HEADER1);
	RH_RF22_systemState = spi_read_reg(RH_RF22_REG_4A_RECEIVED_HEADER0);
	
	// Read the packet
	RH_RF22_packetLength = spi_read_reg(RH_RF22_REG_4B_RECEIVED_PACKET_LENGTH);
	
	// Check packet length and read data
	if (RH_RF22_packetLength <= 40) {
		spiBurstRead(RH_RF22_REG_7F_FIFO_ACCESS, dataIn, RH_RF22_packetLength);
	}
	
	return true;
}

// Wait for packet to be sent, sleep for 16ms until the interrupt wakes us up
uint8_t RH_RF22_wait_for_transmit(void) {
	PCMSK0 |= (1<<nIRQ_INT); // Enable interrupt
	watchdog_sleep(T16MS);
	PCMSK0 &= ~(1<<nIRQ_INT); // Disable interrupt
	RH_RF22_setModeStandby(); // Set to standby as quick as possible
	
	// Check to see if packet was successfully sent
	uint8_t readReg = spi_read_reg(RH_RF22_REG_03_INTERRUPT_STATUS1);
	if (readReg & enpksent) {
		return true;
	}
	
	return false;
}

// Transmit a packet with the request type
void send_packet(uint8_t requestType, uint8_t info, uint8_t *packetBuffer, uint8_t packetLength) {
	RH_RF22_resetFifos();
	RH_RF22_clearInterrupts();
	RH_RF22_txHeaderRequestType = requestType;
	RH_RF22_txHeaderInfo = info;
	RH_RF22_send(packetBuffer, packetLength);
	RH_RF22_wait_for_transmit();
}

// Check in to the server to know what our system state should be. Send the server a random number, copy the HMAC result
// the server generated with the system state, copy the system state to our random number and HMAC it with our private key.
// If the HMAC matches what the server gave us then set the received system state
void server_check_in(void) {
	// Generate random number
	atsha204a_command_random_number(randomNumber, DELAY_WDT);
	
	// Check to see how many good/failed check ins to the server we have. Bump up the Si4432 dBm power if we are failing 
	// to check in (negative checkInCounter) and reduce it if we are checking in fine (positive checkInCounter)
	if (checkInCounter <= 10) {
		if (si443OutputPower < RH_RF22_TXPOW_8DBM) {
			si443OutputPower++;
			RH_RF22_setTxPower(si443OutputPower);
		}
		checkInCounter = 0; // Reset
	}
	else if (checkInCounter >= 120) { // It takes a lot of successful requests to reduce our transmit power
		if (si443OutputPower > RH_RF22_TXPOW_1DBM) {
			si443OutputPower--;
			RH_RF22_setTxPower(si443OutputPower);
		}
		checkInCounter = 0; // Reset
	}
	
	// Send random number
	send_packet(CHECK_IN, (batteryVoltage / 4), randomNumber, 32);
	
	// Wait for server to do nonce and HMAC
	watchdog_sleep(T128MS);
	watchdog_sleep(T32MS);
	//watchdog_sleep(T16MS);
	
	// Wait for packet to match our address
	if (wait_for_valid_packet(WAIT_32MS)) {
		
		// Set alarm state given to us by server to our random number
		/*randomNumber[0] = dataIn[0];
		
		// Generate HMAC from the random number we generated and the alarm state
		atsha204a_command_nonce(randomNumber, DELAY_WDT);
		atsha204a_command_hmac(DELAY_WDT);
		
		// Compare HMACs (31 bytes, elements 1 to 31, dataBuffer[1] ignored)
		uint8_t compareMatched = true;
		for (uint8_t c = 1; c < 31; c++) {
			if (dataBuffer[c+1] != dataIn[c]) {
				compareMatched = false;
			}
		}*/
		
		// To verify sensor number/system state hasn't changed, XOR our sensor number with the first byte of the random number
		// and XOR the system state received with the second byte
		randomNumber[0] ^= RH_SENSOR_NO;
		randomNumber[1] ^= RH_RF22_systemState;
		
		// Generate HMAC from the random number we generated (which includes XOR'd sensor number/system state)
		atsha204a_command_nonce(randomNumber, DELAY_WDT);
		atsha204a_command_hmac(DELAY_WDT);
		
		// Compare HMACs (32 bytes, dataBuffer[0] ignored as it contains the ATSHA packet length)
		uint8_t compareMatched = true;
		for (uint8_t c = 0; c < 32; c++) {
			if (dataBuffer[c+1] != dataIn[c]) {
				compareMatched = false;
			}
		}
		
		// Update alarm state if compare matched
		if (compareMatched == true) {
			systemStateReceived = RH_RF22_systemState;
			checkInCounter++;
		}
	}
	else {
		checkInCounter--;
	}
}

// Once in a while, validate ourselves to the server by requesting a random number and proving we can HMAC it with the
// shared private key which the server will then verify. This is also used to alert the server if the sensor is triggered
uint8_t validate_sensor_to_server(void) {
	// Request random number from server
	send_packet(REQUEST_RANDOM_NO, 0, randomNumber, 1); // Send 1 byte (doesn't matter what it is)
	
	// Wait for server to generate random number
	watchdog_sleep(T32MS);
	watchdog_sleep(T16MS);
	
	// Wait for packet to match our address
	if (wait_for_valid_packet(WAIT_64MS)) {
		/*dataIn[0] = systemState; // Set our system state so the server knows what it is
		
		// Generate HMAC from random number received
		atsha204a_command_nonce(dataIn, DELAY_WDT);
		atsha204a_command_hmac(DELAY_WDT);
		dataBuffer[1] = systemState; // Set system state again
		*/
		
		// To make sure sensor number/system state can't be changed, XOR our sensor number with the first byte of the random number received
		// and XOR the our system state with the second byte
		dataIn[0] ^= RH_SENSOR_NO;
		dataIn[1] ^= systemState;
		
		// Generate HMAC from random number received
		atsha204a_command_nonce(dataIn, DELAY_WDT);
		atsha204a_command_hmac(DELAY_WDT);
		
		// Transmit HMAC
		send_packet(SENSOR_VALIDATE, systemState, &dataBuffer[1], 32); // Starts at [1] as ignoring packet length [0]
		
		// If alarm is triggered, wait to receive confirmation back from server
		if (systemState == ALARM_TRIGGERED) {
			// Wait for server to do nonce and HMAC
			//watchdog_sleep(T128MS);
			//watchdog_sleep(T32MS);
			//watchdog_sleep(T16MS);
			
			// Generate another HMAC from the HMAC we just sent, so we can validate the servers response to acknowledge they received it
			for (uint8_t x = 0; x < 32; x++) {
				randomNumber[x] = dataBuffer[x+1];
			}
			atsha204a_command_nonce(randomNumber, DELAY_WDT);
			atsha204a_command_hmac(DELAY_WDT);
			
			// Wait for packet to match our address
			if (wait_for_valid_packet(WAIT_64MS)) {
				
				// Compare HMACs (32 bytes, dataBuffer[0] ignored as it contains the ATSHA packet length)
				for (uint8_t c = 0; c < 32; c++) {
					if (dataBuffer[c+1] != dataIn[c]) {
						return false; // Compare failed
					}
				}
				
				return true;
			}
		}
	}
	
	return false;
}
