/*
 Alarm System v3.0 Server
 Version: Test
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 1/04/2017
 Last Modified: 1/06/2017
 
 Parts:
 ATmega168 with 12MHz crystal
 ATSHA204A - random numbers and HMAC verification
 Si4432 - 434 MHz radio
 MCP7940M - RTC
 24LC128 - 128Kbit I2C EEPROM (supports 64 byte page write)
 ESP8266 - webserver, serial connection to ATmega
 
 Remember to program your fuses to use an external crystal:
 avrdude -p atmega168p -c usbasp -U lfuse:w:0xef:m -U hfuse:w:0xdf:m -U efuse:w:0xf9:m
 
 */

#define F_CPU 12000000

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <stdio.h>
#include <stdlib.h>
#include "si4432.c"
#include "setup.c"
#include "functions.c" // Application specific functions
#include "atsha204a.c"

int main(void) {
	setup();
	program_rtc_time(); // Program RTC as 0's and start RTC
	
	// Wait for ESP8266 to start up and give wake command 'W'
	while (USART_Receive() != 'W');
	PORTB |= (1<<LED);
	_delay_ms(50);
	PORTB &= ~(1<<LED);
	
	// Send last 50 log items
	send_log();
	_delay_ms(300);
	
	// Send email addresses
	send_email_addresses();
	
	// Enable USART RX interrupt
	UCSR0B |= (1<<RXCIE0);
	
	while(1) {
		// Wait for packet to match our address
		if (wait_for_valid_packet(NO_TIMEOUT)) {
			
			// Client check in
			if (RH_RF22_rxRequestType == CHECK_IN) {
				
				// If some sensor types are enabled, only turn on the sensors with those types
				if (sensorTypeArmed != 0xFF) {
					
					// Read EEPROM to see sensor type
					soft_i2c_eeprom_read_x_bytes(EEPROM_ADDR, ((RH_RF22_fromAddress-1) * 64) + sensorTypeLocation, 1);
					
					// If bits match, set to on if we are armed (eepromBuffer[0] contains the type - 1, 2, 3)
					if (sensorTypeArmed & (1<<(eepromBuffer[0]-1))) {
						systemStateToSend = systemState; // Set state
						if (systemState == SYSTEM_ARMED) { // Set sensor to on if we are armed
							systemStateToSend = SYSTEM_ON;
						}
					}
					else {
						systemStateToSend = SYSTEM_OFF;
					}
				}
				else { // All sensors just match system state
					systemStateToSend = systemState; // Set state
					if (systemState == SYSTEM_ARMED) { // Set sensor to on if we are armed
						systemStateToSend = SYSTEM_ON;
					}
				}
				
				// To make sure sensor number/system state can't be changed, XOR the sensor number with the first byte of the random number
				// and XOR the system state with the second byte
				dataIn[0] ^= RH_RF22_fromAddress;
				dataIn[1] ^= systemStateToSend;
				
				// Generate HMAC from random number received (which includes XOR'd sensor number/system state) and transmit
				atsha204a_command_nonce(dataIn, DELAY_STD);
				atsha204a_command_hmac(DELAY_STD);
				
				// Send a few times to compensate for client's watchdog timer inaccuracy
				for (uint8_t x = 0; x < 5; x++) {
					send_packet(CHECK_IN_REPLY, &dataBuffer[1], 32);
					_delay_ms(5);
				}
				
				// Update ESP8266
				update_sensor_info();
				_delay_ms(25);
			}
			
			// Client requesting random number
			else if (RH_RF22_rxRequestType == REQUEST_RANDOM_NO) {
				
				// Generate random number, transmit random number and wait for client to do nonce and HMAC
				atsha204a_command_random_number(randomNumber, DELAY_STD);
				send_packet(CHECK_IN_REPLY, randomNumber, 32);
				_delay_ms(120);
				
				// Setup Si4432 to only listen for this clients from address
				//spi_write_reg(RH_RF22_REG_32_HEADER_CONTROL1, RH_RF22_HDCH_HEADER3);
				//spi_write_reg(RH_RF22_REG_3F_CHECK_HEADER3, RH_SENSOR_NO | RH_RF22_fromAddress);
				
				if (wait_for_valid_packet(TIMEOUT_250MS)) {
					
					// If sensor is sending alarm triggered, do a nonce and HMAC on the clients incoming HMAC packet
					// and send the result back to confirm we got the packet
					if (RH_RF22_rxInfo == ALARM_TRIGGERED) {
						atsha204a_command_nonce(dataIn, DELAY_STD);
						atsha204a_command_hmac(DELAY_STD);
						_delay_ms(10);
						
						// Send a few times to compensate for client's watchdog timer inaccuracy
						for (uint8_t x = 0; x < 5; x++) {
							send_packet(CHECK_IN_REPLY, &dataBuffer[1], 32);
							_delay_ms(10);
						}
					}
					
					// To verify sensor number/system state hasn't changed, XOR the sensor number with the first byte of the random number
					// and XOR the system state received with the second byte
					randomNumber[0] ^= RH_RF22_fromAddress;
					randomNumber[1] ^= RH_RF22_rxInfo;
					
					// Generate HMAC from the random number we first generated
					atsha204a_command_nonce(randomNumber, DELAY_STD);
					atsha204a_command_hmac(DELAY_STD);
					
					// Compare HMACs (32 bytes, dataBuffer[0] ignored as it contains the ATSHA packet length)
					uint8_t compareMatched = true;
					for (uint8_t c = 0; c < 32; c++) {
						if (dataBuffer[c+1] != dataIn[c]) {
							compareMatched = false;
						}
					}
					
					// Set sensor as valid if compare matched
					if (compareMatched == true) {
						PORTB &= ~(1<<LED);
						
						// Check if alarm is on and if sensor was triggered
						if (systemState == SYSTEM_ON && RH_RF22_rxInfo == ALARM_TRIGGERED) {
							
							// If some sensor types are enabled, only turn on the sensors with those types
							uint8_t sensorTypeCheck = false;
							if (sensorTypeArmed != 0xFF) {
								
								// Read EEPROM to see sensor type
								soft_i2c_eeprom_read_x_bytes(EEPROM_ADDR, ((RH_RF22_fromAddress-1) * 64) + sensorTypeLocation, 1);
								
								// If bits match, set to on if we are armed (eepromBuffer[0] contains the type - 1, 2, 3)
								if (sensorTypeArmed & (1<<(eepromBuffer[0]-1))) {
									sensorTypeCheck = true;
								}
							}
							else {
								sensorTypeCheck = true;
							}
							
							// Passed the sensor type check?
							if (sensorTypeCheck == true) {
								// Sensor address
								char fromAddressStr[5];
								itoa(RH_RF22_fromAddress, fromAddressStr, 10);
								
								datetime_to_buffer();
								strncat(eepromBuffer, "Alarm triggered by S", 21);
								strncat(eepromBuffer, fromAddressStr, 4);
								
								update_log_and_send();
								_delay_ms(25);
								
								// Activate siren if it isn't on
								if (sirenState == SIREN_OFF) {
									sirenState = SIREN_ON;
									
									datetime_to_buffer();
									strncat(eepromBuffer, "Siren on", 9);
									update_log_and_send();
									_delay_ms(25);
									
									// Send to ESP8266
									USART_Transmit('T');
									USART_Transmit('-');
									USART_Transmit('1');
									_delay_ms(25);
								}
							}
						}
						
						update_sensor_verification();
						_delay_ms(25);
					}
				}
				
				// Set Si4432 to listen to any from address
				
			}
		}
		
		// Received serial data from ESP8266
		if (receivedUsart == true) {
			receivedUsart = false;
			
			// Alarm state (e.g A-1)
			if (usartRxBuffer[0] == 'A' && usartRxBuffer[1] == '-') {
				datetime_to_buffer();
				
				if (usartRxBuffer[2] == '1') { // Turn on alarm
					systemState = SYSTEM_ON;
					sirenState = SIREN_OFF;
					strncat(eepromBuffer, "Alarm on", 9);
					update_log_and_send();
					_delay_ms(25);
				}
				else if (usartRxBuffer[2] == '2') { // Turn off alarm
					systemState = SYSTEM_OFF;
					strncat(eepromBuffer, "Alarm off", 10);
					update_log_and_send();
					_delay_ms(25);
					
					if (sirenState == SIREN_ON) {
						sirenState = SIREN_OFF;
						sirenStateCounter = 0;
						datetime_to_buffer();
						strncat(eepromBuffer, "Siren off", 10);
						update_log_and_send();
						_delay_ms(25);
					}
				}
				else if (usartRxBuffer[2] == '3') { // Arm alarm
					systemState = SYSTEM_ARMED;
					sirenState = SIREN_OFF;
					strncat(eepromBuffer, "Alarm armed", 12);
					update_log_and_send();
					_delay_ms(25);
				}
			}
			
			// Update date/time (e.g D-12-3-17-11-23-45-AM-)
			else if (usartRxBuffer[0] == 'D') {
				const char delim[2] = "-";
				uint8_t tokcount = 0;
				
				char *token = strtok(usartRxBuffer, delim);
				while (token != NULL) {
					if (tokcount == 1) {
						rtcYear = atoi(token);
					}
					else if (tokcount == 2) {
						rtcMon = atoi(token);
					}
					else if (tokcount == 3) {
						rtcDate = atoi(token);
					}
					else if (tokcount == 4) {
						rtcHr = atoi(token);
					}
					else if (tokcount == 5) {
						rtcMin = atoi(token);
					}
					else if (tokcount == 6) {
						rtcSec = atoi(token);
					}
					else if (tokcount == 7) {
						rtcAmPm = atoi(token);
					}
					
					tokcount++;
					token = strtok(NULL, delim);
				}
				
				if (tokcount == 8) {
					program_rtc_time();
				}
			}
			
			// Update sensor name (e.g N-1-Kitchen-)
			else if (usartRxBuffer[0] == 'N') {
				const char delim[2] = "-";
				uint8_t tokcount = 0;
				uint8_t sensorNo = 0;
				
				char *token = strtok(usartRxBuffer, delim);
				while (token != NULL) {
					if (tokcount == 1) { // Sensor
						if (atoi(token) <= 0 || atoi(token) >= MAX_SENSORS) { // Out of bounds
							break;
						}
						
						sensorNo = atoi(token) - 1;
					}
					else if (tokcount == 2) { // Name
						strncpy(eepromBuffer, token, 20);
						soft_i2c_eeprom_page_write_64_bytes(EEPROM_ADDR, (sensorNo * 64) + sensorNameLocation);
					}
					
					tokcount++;
					token = strtok(NULL, delim);
				}
			}
			
			// Update sensor type/zone (e.g U-1-PIR-3-)
			else if (usartRxBuffer[0] == 'U') {
				const char delim[2] = "-";
				uint8_t tokcount = 0;
				uint8_t sensorNo = 0;
				
				char *token = strtok(usartRxBuffer, delim);
				while (token != NULL) {
					if (tokcount == 1) { // Sensor
						if (atoi(token) <= 0 || atoi(token) >= MAX_SENSORS) { // Out of bounds
							break;
						}
						
						sensorNo = atoi(token) - 1;
					}
					else if (tokcount == 2) { // Type
						eepromBuffer[0] = atoi(token);
						soft_i2c_eeprom_page_write_64_bytes(EEPROM_ADDR, (sensorNo * 64) + sensorTypeLocation);
					}
					else if (tokcount == 3) { // Zone
						eepromBuffer[0] = atoi(token);
						soft_i2c_eeprom_page_write_64_bytes(EEPROM_ADDR, (sensorNo * 64) + sensorZoneLocation);
					}
					
					tokcount++;
					token = strtok(NULL, delim);
				}
			}
			
			// Update sensor types affected by arming/turning on alarm (e.g. t-1)
			else if (usartRxBuffer[0] == 't' && usartRxBuffer[1] == '-') {
				sensorTypeArmed = usartRxBuffer[2];
			}
			
			// Update email address (e.g E-1-me@example.com-)
			else if (usartRxBuffer[0] == 'E' && usartRxBuffer[1] == '-') {
				const char delim[2] = "-";
				uint8_t tokcount = 0;
				uint8_t emailSlot = 0;
				
				char *token = strtok(usartRxBuffer, delim);
				while (token != NULL) {
					if (tokcount == 1) { // Email slot
						if (atoi(token) <= 0 || atoi(token) >= 3) { // Out of bounds
							break;
						}
						
						emailSlot = atoi(token) - 1;
					}
					else if (tokcount == 2) { // Type
						strncpy(eepromBuffer, token, 50);
						soft_i2c_eeprom_page_write_64_bytes(EEPROM_ADDR, (emailSlot * 64) + emailAddressLocation);
					}
					
					tokcount++;
					token = strtok(NULL, delim);
				}
			}
		}
		
		// Turn off siren after 5 minutes if it was on
		if (sirenStateCounter >= 54) {
			sirenStateCounter = 0;
			sirenState = SIREN_OFF;
			
			datetime_to_buffer();
			strncat(eepromBuffer, "Siren off", 10);
			update_log_and_send();
			_delay_ms(25);
			
			// Send to ESP8266
			USART_Transmit('T');
			USART_Transmit('-');
			USART_Transmit('2');
			_delay_ms(25);
		}
		
		PORTB &= ~(1<<LED);
		_delay_ms(10);
	}
}

ISR(USART_RX_vect) {
	receivedUsart = true;
	
	// Read serial data
	usartRxLength = 0;
	uint8_t timeout = 0;
	while (usartRxLength < 30 && timeout <= 200) {
		usartRxBuffer[usartRxLength] = UDR0;
		usartRxLength++;
		
		timeout = 0;
		while (!(UCSR0A & (1<<RXC0)) && timeout <= 200) { // Wait for data to be received
			_delay_us(10);
			timeout++;
		}
	}
	usartRxBuffer[usartRxLength] = 0;
}

ISR(TIMER1_OVF_vect) {
	// Turn off the siren after 5 minutes
	if (sirenState == SIREN_ON) {
		sirenStateCounter++;
	}
}
